from abc import ABC, abstractmethod

from vistral.command_bus.command import CommandHandler, TCommand


class CommandHandlerResolver(ABC):
    @abstractmethod
    def resolve_command_handler(self, handler_cls: type[CommandHandler[TCommand]], /) -> CommandHandler[TCommand]:
        pass


class DefaultResolver(CommandHandlerResolver):
    def resolve_command_handler(self, handler_cls: type[CommandHandler[TCommand]], /) -> CommandHandler[TCommand]:
        return handler_cls()



from __future__ import annotations

from vistral.command_bus.command import TCommand
from vistral.core.exceptions import VistralError


class CommandAlreadyRegisteredError(VistralError):
    @classmethod
    def for_command(cls, command_cls: type[TCommand], /) -> CommandAlreadyRegisteredError:
        return cls(f"Command handler for {command_cls.__name__} already registered.")


class CommandHandlerNotExists(VistralError):
    @classmethod
    def for_command(cls, command_cls: type[TCommand], /) -> CommandHandlerNotExists:
        return cls(f"Command handler for {command_cls.__name__} does not exist.")



from typing import Optional, cast

try:
    from lagom import Container
    from lagom.container import CallTimeContainerUpdate
except ImportError as import_error:
    raise ImportError(
        f"`lagom` dependency is missing: {import_error}. "
        + "To use LagomResolver, Re-install `vistral` with [lagom] extra dependency."
    ) from None

from vistral.command_bus.command import CommandHandler, TCommand
from vistral.command_bus.resolver import CommandHandlerResolver


class LagomResolver(CommandHandlerResolver):
    def __init__(
        self,
        container: Container,
        shared_deps: Optional[list[type]] = None,
        container_updater: Optional[CallTimeContainerUpdate] = None,
    ):
        self._partial = lambda func: container.partial(
            func=func, shared=shared_deps, container_updater=container_updater
        )

    def resolve_command_handler(
        self,
        handler_cls: type[CommandHandler[TCommand]],
    ) -> CommandHandler[TCommand]:
        return cast(
            CommandHandler[TCommand], self._partial(func=handler_cls)()
        )  # TODO: add validation and runtime error once not all arguments are bound



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, TypeVar


@dataclass(frozen=True)
class Command:
    pass


TCommand = TypeVar("TCommand", bound=Command)


class CommandHandler(Generic[TCommand], ABC):
    @abstractmethod
    def __call__(self, command: TCommand) -> None:
        pass



from vistral.command_bus.command import Command, CommandHandler, TCommand
from vistral.command_bus.exceptions import CommandAlreadyRegisteredError, CommandHandlerNotExists
from vistral.command_bus.resolver import CommandHandlerResolver, DefaultResolver


class CommandBus:
    def __init__(self, resolver: CommandHandlerResolver = DefaultResolver()):
        self._resolver = resolver
        self._handlers: dict[type[Command], type[CommandHandler]] = {}

    def register(
        self,
        command_type: type[TCommand],
        handler_type: type[CommandHandler[TCommand]],
    ) -> None:
        if command_type in self._handlers:
            raise CommandAlreadyRegisteredError.for_command(command_type)

        self._handlers[command_type] = handler_type

    def handle(self, command: TCommand) -> None:
        command_type = type(command)
        try:
            handler_cls = self._handlers[command_type]
        except KeyError:
            raise CommandHandlerNotExists.for_command(command_type)

        handler = self._resolver.resolve_command_handler(handler_cls)
        return handler(command=command)
